using System;
using System.Collections.Specialized;
using System.Collections;
using InfiniTec.SystemUtilities.DataAccess;

namespace InfiniTec.WebDAV
{
	/// <summary>
	/// Summary description for WebDAVElement.
	/// </summary>
	public class WebDAVElement: Element, IWebDAVElement
	{
		protected InfiniTec.WebDAV.WebDAVLayer webDAVLayer;
		private Utilities utilities;
		private InitializationStatus _initializationStatus;
		private bool _RefreshInProgress;

		protected override void OnPropertyAdd(InfiniTec.SystemUtilities.DataAccess.PropertyChangeEventArgs e)
		{
			base.OnPropertyAdd (e);
			utilities.HandlePropertyAdd(e, _initializationStatus);
		}

		public void ForceFlush(bool onlyDirty) 
		{
			utilities.ForceFlush(onlyDirty);
		}

		public void RetrieveProperties(params string[] propertiesToLoad) 
		{
			IElement details;

			details = webDAVLayer.PropFind(Uri, propertiesToLoad);
			utilities.UpdateProperties(this, details);
		}

		public virtual bool RefreshInProgress 
		{
			get 
			{
				return _RefreshInProgress;
			}
		}

		protected override void OnPropertyChange(InfiniTec.SystemUtilities.DataAccess.PropertyChangeEventArgs e)
		{
			string oldName;

			base.OnPropertyChange (e);
			utilities.HandlePropertyChange(e, _initializationStatus);
			if (e.Property.Name == "DAV:displayname") 
			{
				Refresh();
				oldName = Uri.Substring(Uri.LastIndexOf("/")+1);
				_Uri = (string) Properties["DAV:href"].Value;
				OnRenamed(new ElementRenamedEventArgs(oldName));
			}
		}
		public void Remove() 
		{
			utilities.Remove();
			OnDeleted();
		}

		internal protected void Initialize()
		{
			if (_initializationStatus == InitializationStatus.None) 
			{
				Refresh();
			}
		}

		public bool CompleteInitialization() 
		{
			if (_initializationStatus != InitializationStatus.Full) 
			{
				Refresh();
				_initializationStatus = InitializationStatus.Full;
				return true;
			}
			else return false;
		}

		public virtual void Refresh() 
		{
			IElement element;

		
			if (!_RefreshInProgress) 
			{
				_RefreshInProgress = true;
				try 
				{
					lock (webDAVLayer) 
					{
						element = webDAVLayer.PropFind(Uri);
						utilities.UpdateProperties(this, element);
					}
				}
				finally 
				{
					_RefreshInProgress = false;
				}
			}
		}

		public virtual void LoadProperties(params string[] propertiesToLoad) {
			IElement element;

			lock (webDAVLayer) 
			{
				element = webDAVLayer.PropFind(Uri, propertiesToLoad);
			}
			utilities.UpdateProperties(this, element);
		}


		public override PropertyCollection Properties 
		{
			get 
			{
				Initialize();
				return base.Properties;
			}
		}
		public WebDAVElement(string uri, Connection connection): this(uri, connection, InitializationStatus.None)
		{
		}
		protected internal WebDAVElement(string uri, Connection connection, InitializationStatus initializationStatus): base(uri)
		{
			_RefreshInProgress = false;

			webDAVLayer = new WebDAVLayer(new WebDAVObjectFactory(connection));
			webDAVLayer.Credential = connection.Credential;
			webDAVLayer.ConnectionGroup = connection.GetHashCode().ToString();

			utilities = new Utilities(this, webDAVLayer);
			_initializationStatus = initializationStatus;
		}

		public NameValueCollection Capabilities() 
		{
			return utilities.Capabilities(this.Uri);
		}

		#region IWebDAVElement Members

		public void Move(IWebDAVFolder destination)
		{
			lock (webDAVLayer) 
			{
				webDAVLayer.Move(destination, this);
			}
		}

		public void Copy(IWebDAVFolder destination)
		{
			lock (webDAVLayer) 
			{
				webDAVLayer.Copy(destination, this);
			}
		}

		public void Rename(string newName)
		{
			Properties["DAV:displayName"].Value = newName;
			Refresh();
			_Uri = (string) Properties["DAV:href"].Value;
		}

		public void Delete()
		{
			webDAVLayer.Delete(this);
		}

		#endregion
	}
}
