using System;
using System.Threading;
using System.Collections.Specialized;
using System.Collections;
using InfiniTec.SystemUtilities.DataAccess;

namespace InfiniTec.WebDAV
{
	/// <summary>
	/// Summary description for WebDAVFolder.
	/// </summary>
	public class WebDAVFolder: Folder, IWebDAVFolder
	{
		public event StatusChangedEventHandler StatusChanged;

		protected InfiniTec.WebDAV.WebDAVLayer webDAVLayer;
		private Utilities utilities;
		private InitializationStatus _initializationStatus;
		private bool _RefreshInProgress;

		public virtual bool RefreshInProgress 
		{
			get 
			{
				return _RefreshInProgress;
			}
		}

		protected override void OnPropertyAdd(InfiniTec.SystemUtilities.DataAccess.PropertyChangeEventArgs e)
		{
			base.OnPropertyAdd (e);

			utilities.HandlePropertyAdd(e, _initializationStatus);
		}

		public void ForceFlush(bool onlyDirty) 
		{
			utilities.ForceFlush(onlyDirty);
		}

		public NameValueCollection Capabilities() 
		{
			return utilities.Capabilities(this.Uri);
		}

		protected void OnStatusChanged(StatusChangedEventArgs e) 
		{
			if (StatusChanged != null) StatusChanged(this, e);
		}
		protected override void OnPropertyChange(InfiniTec.SystemUtilities.DataAccess.PropertyChangeEventArgs e)
		{
			string oldName;

			base.OnPropertyChange (e);
			utilities.HandlePropertyChange(e, _initializationStatus);
			if (e.Property.Name == "DAV:displayname") 
			{
				Refresh();
				oldName = Uri.Substring(Uri.LastIndexOf("/")+1);
				_Uri = (string) Properties["DAV:href"].Value;
				OnRenamed(new ElementRenamedEventArgs(oldName));
			}
		}

		public WebDAVFolder(string uri, Connection connection): this(uri, connection, InitializationStatus.None)
		{
		}

		protected internal WebDAVFolder(string uri, Connection connection, InitializationStatus initializationStatus): base(uri)
		{
			_RefreshInProgress = false;

			webDAVLayer = new InfiniTec.WebDAV.WebDAVLayer(new WebDAVObjectFactory(connection));
			webDAVLayer.Credential = connection.Credential;
			webDAVLayer.ConnectionGroup = connection.GetHashCode().ToString();

			utilities = new Utilities(this, webDAVLayer);

			_initializationStatus = initializationStatus;
		}
		public virtual void Rename(string newName) 
		{
			utilities.Rename(newName);
		}

		internal protected virtual void Initialize()
		{
			if (_initializationStatus == InitializationStatus.None) 
			{
				Refresh();
			}
		}

		public void Remove() 
		{
			utilities.Remove();
			OnDeleted();
		}
		public void Remove(params IElement[] elements) 
		{
			string relativeUri;

			lock (webDAVLayer) 
			{
				webDAVLayer.Delete(elements);
			}
			foreach (IElement element in elements) 
			{
				relativeUri = element.Uri.Substring(element.Uri.LastIndexOf("/")+1);
				Elements.Remove(relativeUri);
			}
		}

		public virtual bool CompleteInitialization() 
		{
			if (_initializationStatus != InitializationStatus.Full) 
			{
				_initializationStatus = InitializationStatus.Full;
				Refresh();
				return true;
			}
			return false;
		}

		protected virtual void MergeElements(IElement[] elements, IDictionary destination) 
		{
			string relativeName;

			foreach (Element element in elements) 
			{
				relativeName = element.Uri;
				relativeName = relativeName.Substring(relativeName.LastIndexOf("/")+1).ToLower();
				if (!destination.Contains(relativeName)) 
				{
					destination.Add(relativeName, element);
					element.Deleted += new EventHandler(element_Deleted);
					element.Renamed += new ElementRenamedEventHandler(element_Renamed);
				}
				else 
				{
					utilities.UpdateProperties((IElement) destination[relativeName], element);
				}
			}
		}

		public virtual void RemoveExistingElements(IElement[] elements, IDictionary destination) 
		{
			string relativeName;

			foreach (Element element in elements) 
			{
				relativeName = element.Uri;
				relativeName = relativeName.Substring(relativeName.LastIndexOf("/")+1).ToLower();
				if (destination.Contains(relativeName)) destination.Remove(relativeName);
			}
		}

		public virtual void Refresh() 
		{
			Refresh("DAV:displayname", "DAV:contentclass", "DAV:getcontentlength",
				"DAV:isfolder", "DAV:hassubs", "urn:schemas:contacts:cn", "DAV:ishidden",
				"DAV:getlastmodified", "DAV:hasattachment");
		}

		public virtual IWebDAVElement[] Search(string conditions, params string[] propertiesToLoad) 
		{
			const int pageSize = 50;


			int position;
			ArrayList foundElements;
			IElement[] elements;

			foundElements = new ArrayList();

			position = 0;

			do 
			{
				elements = webDAVLayer.Search(Uri, string.Format("{0}-{1}", position, position+pageSize-1), conditions, propertiesToLoad);
				foreach (IElement element in elements) 
				{
					foundElements.Add(element);
				}
				position += pageSize;
			}
			while (elements.Length == pageSize);

			return (IWebDAVElement[]) foundElements.ToArray(typeof(IWebDAVElement));
		}

		public void RetrieveProperties(params string[] propertiesToLoad) 
		{
			IElement details;

			details = webDAVLayer.PropFind(Uri, propertiesToLoad);
			utilities.UpdateProperties(this, details);
		}

		protected virtual void Refresh(params string[] propertiesToLoad) 
		{
			const int pageSize = 50;

			IElement[] elements;
			IElement details;
			IDictionary existingChildren;
			int childCount;
			int position;

			if (!_RefreshInProgress) 
			{
				_RefreshInProgress = true;
				try 
				{
					existingChildren = (IDictionary) ((ICloneable) Elements).Clone();

					lock (webDAVLayer) 
					{
						details = webDAVLayer.PropFind(Uri);
						utilities.UpdateProperties(this, details);
						if (_initializationStatus == InitializationStatus.Full) 
						{
							try 
							{
								childCount = (int) details.Properties["DAV:childcount"].Value;
							}
							catch 
							{
								childCount = -1;
							}
						
							position = 0;
							do 
							{
								elements = webDAVLayer.Search(Uri, string.Format("{0}-{1}", position, position+pageSize-1), "", propertiesToLoad);

								MergeElements(elements, Elements);
								RemoveExistingElements(elements, existingChildren);
								position += pageSize;
								OnStatusChanged(new StatusChangedEventArgs((childCount > 0) ? position * 100 / childCount : 0));
							}
							while (elements.Length == pageSize);

							foreach (DictionaryEntry entry in existingChildren) 
							{
								Elements.Remove(entry.Key);
							}
						}
					}
				}
				finally 
				{
					_RefreshInProgress = false;
				}
			}
		}

		public override PropertyCollection Properties 
		{
			get 
			{
				Initialize();
				return base.Properties;
			}
		}
		public override IDictionary Elements 
		{
			get 
			{
				Initialize();
				return base.Elements;
			}
		}
		public virtual void LoadProperties(params string[] propertiesToLoad) 
		{
			IElement element;

			lock (webDAVLayer) 
			{
				element = webDAVLayer.PropFind(Uri, propertiesToLoad);
			}
			utilities.UpdateProperties(this, element);
		}

		public void CreateFolder(string name) 
		{
			IFolder folder;

			if (!Elements.Contains(name.ToLower())) 
			{
				lock (webDAVLayer) 
				{
					folder = webDAVLayer.MkCol(this, name);
				}
				Elements.Add(name.ToLower(), folder);
			}
		}
		#region IWebDAVFolder Members

		public void Move(IWebDAVFolder destination, params IWebDAVElement[] elements)
		{
			lock (webDAVLayer) 
			{
				webDAVLayer.StatusChanged += new StatusChangedEventHandler(webDAVLayer_StatusChanged);
				webDAVLayer.Move(destination, elements);
				webDAVLayer.StatusChanged -= new StatusChangedEventHandler(webDAVLayer_StatusChanged);
			}
		}

		public void Copy(IWebDAVFolder destination, params IWebDAVElement[] elements)
		{
			webDAVLayer.StatusChanged += new StatusChangedEventHandler(webDAVLayer_StatusChanged);
			webDAVLayer.Copy(destination, elements);
			webDAVLayer.StatusChanged -= new StatusChangedEventHandler(webDAVLayer_StatusChanged);
		}

		public void Delete(params IWebDAVElement[] elements)
		{
			// TODO:  Add WebDAVFolder.Delete implementation
		}

		#endregion

		#region IWebDAVElement Members

		void InfiniTec.WebDAV.IWebDAVElement.Move(IWebDAVFolder destination)
		{
			lock (webDAVLayer) 
			{
				webDAVLayer.Move(destination, this);
			}
		}

		void InfiniTec.WebDAV.IWebDAVElement.Copy(IWebDAVFolder destination)
		{
			lock (webDAVLayer) 
			{
				webDAVLayer.Copy(destination, this);
			}
		}

		void InfiniTec.WebDAV.IWebDAVElement.Delete()
		{
			// TODO:  Add WebDAVFolder.InfiniTec.WebDAV.IWebDAVElement.Delete implementation
		}

		#endregion

		private void element_Deleted(object sender, EventArgs e)
		{
			IElement element;
			element = sender as IElement;

			if (element != null) 
			{
				if (Elements.Contains(element.Uri)) Elements.Remove(element.Uri);
			}
		}

		private void element_Renamed(object sender, ElementRenamedEventArgs e)
		{
			IElement element;

			if (Elements.Contains(e.OldName)) 
			{
				element = (IElement) Elements[e.OldName];
				Elements.Remove(element);
				Elements.Add(element.Uri, element);
			}
		}

		private void webDAVLayer_StatusChanged(object sender, StatusChangedEventArgs e)
		{
			OnStatusChanged(e);
		}
	}
}
