using System;
using System.Text;
using System.IO;
using System.Net;
using System.Collections;
using System.Collections.Specialized;
using System.Xml;
using System.Xml.XPath;
using InfiniTec.SystemUtilities.DataAccess;

namespace InfiniTec.WebDAV
{
	/// <summary>
	/// Summary description for Class1.
	/// </summary>
	public class WebDAVLayer
	{
		public static event CertificateProblemOccuredEventHandler CertificateProblemOccured 
		{
			add 
			{
				CertificatePolicy.CertificateProblemOccured += value; 
			}
			remove 
			{
				CertificatePolicy.CertificateProblemOccured -= value;
			}
		}

		private Credential _Credential;

		public event StatusChangedEventHandler StatusChanged;
		private string _ConnectionGroup;
		private WebRequest webRequest;
		private ObjectFactory _objectFactory;

		protected void OnStatusChanged(StatusChangedEventArgs e) 
		{
			if (StatusChanged != null) StatusChanged(this, e);
		}

		public Credential Credential 
		{
			get 
			{
				return _Credential; 
			}
			set 
			{
				_Credential = value;
			}
		}
			
		public ObjectFactory ObjectFactory 
		{
			get { return _objectFactory; }
			set { _objectFactory = value; }
		}

		public string ConnectionGroup 
		{
			get { return _ConnectionGroup; }
			set { _ConnectionGroup = (value != null) ? value : ""; }
		}


		public WebDAVLayer(ObjectFactory objectFactory)
		{
			webRequest = new WebRequest(ConnectionGroup);
			_objectFactory = objectFactory;
			Credential = ((WebDAVObjectFactory) objectFactory).Connection.Credential;
		}

		public WebDAVLayer(Connection connection): this(new WebDAVObjectFactory(connection))
		{
		}

		public IElement PropFind(string uri) 
		{
			return PropFind(uri, 0)[0];
		}

		public IElement[] Search(string uri, string range, string conditions) 
		{
			return Search(uri, range, conditions, null);
		}

		public IElement[] Search(string uri, string range, string conditions, params string[] propertiesToLoad) 
		{
			return Search(uri, range, false, conditions, propertiesToLoad);
		}

		public IElement[] Search(string uri, string range, bool deletedItems, string conditions, params string[] propertiesToLoad) 
		{
			XmlDocument doc;
			XmlNamespaceManager manager;
			XmlNodeList list;
			IElement[] result;
			int i;
			string sql;
			StringBuilder propertyString;
			NameValueCollection customHeader;

			doc = new XmlDocument();
			doc.AppendChild(doc.CreateProcessingInstruction("xml","version=\"1.0\""));
			doc.AppendChild(doc.CreateElement("a:searchrequest", "DAV:"));
			doc.DocumentElement.AppendChild(doc.CreateElement("a:sql", "DAV:"));

			propertyString = new StringBuilder();

			if (propertiesToLoad != null) 
			{
				foreach (string property in propertiesToLoad) 
				{
					propertyString.Append("\"");
					propertyString.Append(property);
					propertyString.Append("\",");
				}
				propertyString.Remove(propertyString.Length-1, 1);
			}
			else 
				propertyString.Append("*");

			customHeader = new NameValueCollection();
			customHeader.Add("Range", "rows=" + range);

			sql = string.Format("SELECT {0} FROM SCOPE('{1} TRAVERSAL OF \"{2}\"')", propertyString.ToString(), (deletedItems) ? "SOFTDELETED" : "SHALLOW", uri);

			if ((conditions != "") && (conditions != null))
			{
				sql += " WHERE " + conditions;
			}

			doc.DocumentElement.FirstChild.AppendChild(doc.CreateTextNode(sql));
			lock (webRequest) 
			{
				doc = webRequest.Request(uri, "SEARCH", customHeader, doc, Credential);
			}

			manager = new XmlNamespaceManager(doc.NameTable);
			manager.AddNamespace("a", "DAV:");

			list = doc.SelectNodes("//a:response", manager);

			result = new IElement[list.Count];

			for (i=0; i<list.Count; i++) 
			{
				result[i] = ExtractElement((XmlElement) list[i]);
			}
			return result;			
		}

		public IElement[] PropFind(string uri, int depth) 
		{
			XmlDocument doc;
			XmlNamespaceManager manager;
			XmlNodeList list;
			IElement[] result;
			int i;

			doc = new XmlDocument();
			doc.AppendChild(doc.CreateProcessingInstruction("xml","version=\"1.0\""));
			doc.AppendChild(doc.CreateElement("a:propfind", "DAV:"));
			doc.DocumentElement.AppendChild(doc.CreateElement("a:allprop", "DAV:"));
			doc.DocumentElement.FirstChild.AppendChild(doc.CreateElement("e:allprop", "http://schemas.microsoft.com/exchange/"));

			lock (webRequest) 
			{
				doc = webRequest.Request(uri, "PROPFIND", doc, depth, Credential);
			}

			manager = new XmlNamespaceManager(doc.NameTable);
			manager.AddNamespace("a", "DAV:");

			list = doc.SelectNodes("//a:response", manager);

			result = new IElement[list.Count];

			for (i=0; i<list.Count; i++) 
			{
				result[i] = ExtractElement((XmlElement) list[i]);
			}
			return result;
		}

		public IElement ExtractElement(XmlElement node) 
		{
			IElement result;
			XmlNamespaceManager manager;
			XmlElement element;
			XmlNodeList list;
			IProperty property;
			string uri;

			manager = new XmlNamespaceManager(node.OwnerDocument.NameTable);
			manager.AddNamespace("a", "DAV:");

			//element = (XmlElement) node.SelectSingleNode("a:propstat[a:status = 'HTTP/1.1 200 OK']/a:prop/a:href", manager);
			element = (XmlElement) node.SelectSingleNode("a:href", manager);
			
			uri = element.InnerText;

			try 
			{
				element = (XmlElement) node.SelectSingleNode("a:propstat[a:status = 'HTTP/1.1 200 OK']/a:prop/a:isfolder", manager);

				if (element.InnerText == "0") 
				{
					result = _objectFactory.CreateElement(uri);
				}
				else 
				{
					result = _objectFactory.CreateFolder(uri);
				}
			} 
			catch 
			{
				result = _objectFactory.CreateElement(uri);
			}	
			list = node.SelectNodes("a:propstat[a:status = 'HTTP/1.1 200 OK']/a:prop/*", manager);

			result.Properties.SuppressEvents = true;
			try 
			{
				foreach (XmlElement elem in list) 
				{
					property = DecodeProperty(elem);
					result.Properties.Add(property);
				}	
			}
			finally 
			{
				result.Properties.SuppressEvents = false;
			}

			return result;
		}
		public XmlDocument PropFindRaw(string uri, params string[] propertiesToLoad) 
		{
			XmlDocument doc;
			XmlElement node;
			string prefix, suffix;

			doc = new XmlDocument();
			doc.AppendChild(doc.CreateProcessingInstruction("xml","version=\"1.0\""));
			doc.AppendChild(doc.CreateElement("a:propfind", "DAV:"));
			doc.DocumentElement.AppendChild(doc.CreateElement("a:prop", "DAV:"));
			
			foreach (string property in propertiesToLoad) 
			{
				prefix = property.Substring(0, property.LastIndexOfAny(new char[] {'/', ':', '#'})+1);
				suffix = property.Substring(property.LastIndexOfAny(new char[] {'/', ':', '#'})+1);
				node = doc.CreateElement(suffix, prefix);
				doc.DocumentElement.FirstChild.AppendChild(node);
			}

			lock (webRequest) 
			{
				doc = webRequest.Request(uri, "PROPFIND", doc, 0, Credential);
			}
			return doc;
		}

		public IElement PropFind(string uri, params string[] propertiesToLoad) 
		{
			XmlDocument doc;
			XmlElement node;
			XmlNamespaceManager manager;

			doc = PropFindRaw(uri, propertiesToLoad);

			manager = new XmlNamespaceManager(doc.NameTable);
			manager.AddNamespace("a", "DAV:");

			node = (XmlElement) doc.SelectSingleNode("//a:response", manager);

			return ExtractElement(node);
		}
		public IElement[] PropFind(string[] uri) 
		{
			return null;
		}
		public IElement[] PropFind(string[] uri, params string[] propertiesToLoad) 
		{
			return null;
		}

		public XmlElement PreparePropPatchRequest() 
		{
			XmlDocument doc;
			XmlElement node;

			doc = new XmlDocument();
			doc.AppendChild(doc.CreateProcessingInstruction("xml","version=\"1.0\""));
			doc.AppendChild(doc.CreateElement("a:propertyupdate", "DAV:"));
			doc.DocumentElement.AppendChild(doc.CreateElement("a:set", "DAV:"));
			node = doc.CreateElement("a:prop", "DAV:");
			return (XmlElement) doc.DocumentElement.FirstChild.AppendChild(node);
		}

		public XmlDocument SendPropPatchRequest(IElement element, XmlElement node) 
		{
			XmlDocument doc;
			lock (webRequest) 
			{
				doc = webRequest.Request(element.Uri, "PROPPATCH", node.OwnerDocument, Credential);
			}			
			return doc;
		}

		public void PropPatch(IElement element, bool onlyDirty) 
		{ 
			XmlElement node;
			int count;

			node = PreparePropPatchRequest();
			count = 0;
			foreach (Property property in element.Properties) 
			{
				if (property.Changed | !onlyDirty) 
				{
					node.AppendChild(EncodeProperty(property, node.OwnerDocument));
					count++;
				}
			}

			if (!onlyDirty || (count > 0)) 
			{
				SendPropPatchRequest(element, node);
			}

			foreach (Property property in element.Properties) 
			{
				property.ResetChangedFlag();
			}
		}

		public Response[] BCopy(string uri, string[] source, string[] dest) 
		{
			XmlDocument doc;
			XmlElement element;
			XmlElement target;
			int i;
			NameValueCollection headers;
			Response[] responses;

			doc = new XmlDocument();

			doc.AppendChild(doc.CreateXmlDeclaration("1.0", "", ""));
			doc.AppendChild(doc.CreateElement("a:copy", "DAV:"));

			for (i=0; i < source.Length; i++) 
			{
				target = doc.CreateElement("a:target", "DAV:");

				element = doc.CreateElement("a:href", "DAV:");
				element.InnerText = source[i];
				target.AppendChild(element);

				element = doc.CreateElement("a:dest", "DAV:");
				element.InnerText = dest[i];
				target.AppendChild(element);

				doc.DocumentElement.AppendChild(target);
			}
			headers = new NameValueCollection();

			headers.Add("Translate", "f");
			lock (webRequest) 
			{
				webRequest.Request(uri, "BCOPY", headers, doc, Credential);
				responses = webRequest.Responses;
			}
			return responses;
		}
		public Response[] BMove(string uri, string[] source, string[] dest) 
		{
			XmlDocument doc;
			XmlElement element;
			XmlElement target;
			int i;
			NameValueCollection headers;
			Response[] responses;

			doc = new XmlDocument();

			doc.AppendChild(doc.CreateXmlDeclaration("1.0", "", ""));
			doc.AppendChild(doc.CreateElement("a:move", "DAV:"));

			for (i=0; i < source.Length; i++) 
			{
				target = doc.CreateElement("a:target", "DAV:");

				element = doc.CreateElement("a:href", "DAV:");
				element.InnerText = source[i];
				target.AppendChild(element);

				element = doc.CreateElement("a:dest", "DAV:");
				element.InnerText = dest[i];
				target.AppendChild(element);

				doc.DocumentElement.AppendChild(target);
			}
			headers = new NameValueCollection();

			headers.Add("Translate", "f");
			lock (webRequest) 
			{
				webRequest.Request(uri, "BMOVE", headers, doc, Credential);
				responses = webRequest.Responses;
			}
			return responses;
		}
		public Response[] BDelete(string uri, string[] elements) 
		{
			XmlDocument doc;
			XmlElement element;
			XmlElement target;
			int i;
			NameValueCollection headers;
			Response[] responses;

			doc = new XmlDocument();

			doc.AppendChild(doc.CreateXmlDeclaration("1.0", "UTF-8", ""));
			doc.AppendChild(doc.CreateElement("a:delete", "DAV:"));

			for (i=0; i < elements.Length; i++) 
			{
				target = doc.CreateElement("a:target", "DAV:");

				element = doc.CreateElement("a:href", "DAV:");
				element.InnerText = elements[i];
				target.AppendChild(element);

				doc.DocumentElement.AppendChild(target);
			}
			headers = new NameValueCollection();

			headers.Add("Translate", "f");
			lock (webRequest) 
			{
				webRequest.Request(uri, "BDELETE", headers, doc, Credential);
				responses = webRequest.Responses;
			}
			return responses;
		}

		public void Copy(IWebDAVFolder destination, params IWebDAVElement[] items) 
		{
			NameValueCollection headers;
			int counter;

			headers = new NameValueCollection();
			headers.Add("Overwrite", "F");
			headers.Add("allow-rename", "t");
			headers.Add("translate", "F");

			counter = 0;
			lock (webRequest) 
			{
				foreach (IWebDAVElement item in items) 
				{ 
					headers.Set("Destination", string.Concat(destination.Uri, "/", (string) item.Properties["DAV:displayname"].Value));

					OnStatusChanged(new StatusChangedEventArgs(item, (counter*100)/items.Length));
					webRequest.Request(item.Uri, "COPY", headers, Credential);
					counter++;
				}
			}
		}

		public void Rename(IWebDAVElement element, string newName) 
		{
			string basePath;
			NameValueCollection headers;

			basePath = element.Uri;
			basePath = basePath.Substring(0, basePath.LastIndexOf("/"));

			headers = new NameValueCollection();
			headers.Add("Overwrite", "f");
			headers.Add("Destination", string.Concat(basePath, "/", newName));

			lock (webRequest) 
			{
				webRequest.Request(element.Uri, "MOVE", headers, Credential);
			}
		}

		public void Move(IWebDAVFolder destination, params IWebDAVElement[] items) 
		{
			NameValueCollection headers;
			int counter;

			headers = new NameValueCollection();
			headers.Add("Overwrite", "t");
			
			counter = 0;
			lock (webRequest) 
			{
				foreach (IWebDAVElement item in items) 
				{
					headers.Set("Destination", string.Concat(destination.Uri, "/", (string) item.Properties["DAV:displayname"].Value));

					OnStatusChanged(new StatusChangedEventArgs(item, (counter*100)/items.Length));
					webRequest.Request(item.Uri, "MOVE", headers, Credential);
					counter++;
				}
			}
		}

		public void Delete(params IElement[] elements) 
		{
			webRequest = new WebRequest(_ConnectionGroup);
			foreach (IElement element in elements) 
			{
				webRequest.Request(element.Uri, "DELETE", Credential);
			}
		}

		public IFolder MkCol(IElement element, string name) 
		{
			return MkCol(element.Uri, name);
		}
			
		public IFolder MkCol(string baseUri, string name) 
		{
			if (!baseUri.EndsWith("/")) baseUri += "/";
			MkCol(baseUri + name);
			return (IFolder) PropFind(baseUri + name);
		}

		public void  MkCol(string uri) {
			WebRequest webRequest;

			webRequest = new WebRequest(_ConnectionGroup);
			webRequest.Request(uri, "MKCOL", Credential);

		}
		public void  MkCol(string uri, IProperty[] properties) 
		{
			WebRequest webRequest;
			XmlElement node;

			webRequest = new WebRequest(_ConnectionGroup);

			node = PreparePropPatchRequest();

			foreach (IProperty property in properties) 
			{
				node.AppendChild(EncodeProperty(property, node.OwnerDocument));
			}

			webRequest.Request(uri, "MKCOL", node.OwnerDocument, Credential);
		}

		private static object DecodeValue(string value, string datatype) 
		{
			if (datatype.StartsWith("mv.")) datatype = datatype.Substring(3);
			switch (datatype) 
			{
				case "boolean": return XmlConvert.ToBoolean(value);
				case "dateTime.tz": return XmlConvert.ToDateTime(value);
				case "i2": return XmlConvert.ToInt16(value);
				case "int": return XmlConvert.ToInt32(value);
				case "i8": return XmlConvert.ToInt64(value);
				case "r4": return XmlConvert.ToSingle(value);
				case "fixed.14.4": return XmlConvert.ToDouble(value);
				case "float": return XmlConvert.ToSingle(value);
				case "uuid": return XmlConvert.ToGuid(value);
				case "bin.base64": return Convert.FromBase64String(value);
				default: return value;
			}
		}

		private static string EncodeValue(object value, string dataType) 
		{
			Type type;
			string result;

			result = null;

			type = value.GetType();

			if (dataType.IndexOf("boolean") != -1) 
			{
				result = (bool) value ? "1" : "0";
			}
			else if (dataType.IndexOf("bin.base64") != -1) 
			{
				result = Convert.ToBase64String((byte[]) value);
			}
			else if (dataType.IndexOf("datetime.tz") != -1) 
			{
				result = ((DateTime) value).ToString("yyyy-MM-ddTHH:mm:sszzz");
			}
			else result = value.ToString();

			return result;
		}

		public static IProperty DecodeProperty(XmlNode node) 
		{
			XmlAttribute attribute;
			ArrayList list;
			XmlDocument doc;
			string prefix;
			string datatype;
			object value;

			doc = node.OwnerDocument;
			prefix = node.GetPrefixOfNamespace("urn:uuid:c2f41010-65b3-11d1-a29f-00aa00c14882/");
			attribute = node.Attributes[prefix + ":dt"];
			if (attribute == null) datatype = "string";
			else datatype = attribute.Value;

			if (datatype.StartsWith("mv")) 
			{
				list = new ArrayList();
				foreach (XmlNode childnode in node.ChildNodes) 
				{
					if (childnode.NamespaceURI == "xml:") 
					{
						list.Add(DecodeValue(childnode.InnerText, datatype));
					}
				}
				value = list;
			}
			else value = DecodeValue(node.InnerText, datatype);

			prefix = node.NamespaceURI;

			if ((!prefix.EndsWith("/"))  && (!prefix.EndsWith(":"))) prefix += "#";
			return (IProperty) new Property(prefix + node.LocalName, datatype, value);
		}

		public static XmlNode EncodeProperty(IProperty property, XmlDocument doc) 
		{
			XmlElement node;
			XmlElement subNode;
			string name;
			string prefix;

			lock (doc)
			{
				prefix = property.Name.Substring(0, property.Name.LastIndexOfAny(new char[] {'/', ':', '#'})+1);
				if (prefix.EndsWith("#")) prefix = prefix.Substring(0, prefix.Length-1);
				name = property.Name.Substring(property.Name.LastIndexOfAny(new char[] {'/', ':', '#'})+1);
				node = doc.CreateElement(name, prefix);
			}
			if (property.DataType.StartsWith("mv")) 
			{
				foreach (object v in (ArrayList) property.Value) 
				{
					subNode = doc.CreateElement("v", "xml:");
					subNode.InnerText = EncodeValue(v, property.DataType);
					node.AppendChild(subNode);
				}
			} 
			else 
			{
				node.InnerText = EncodeValue(property.Value, property.DataType);
			}
			node.SetAttribute("dt", "urn:uuid:c2f41010-65b3-11d1-a29f-00aa00c14882/", property.DataType);

			return node;
		}

		public XmlDocument X_MS_EnumAtts(IWebDAVElement element)
		{
			XmlDocument doc;
			WebRequest request;

			request = new WebRequest(_ConnectionGroup);
			lock (request) 
			{
				doc = request.Request(element.Uri, "X-MS-ENUMATTS", Credential);
			}
			return doc;
		}

		public void Get(string uri, Stream stream) 
		{
			const int size = 65536;
			byte[] bytes;
			int numBytes;
			HttpWebRequest request;
			WebResponse response;
			Stream responseStream;

			bytes = new byte[size];
			request = WebRequest.GetRequestObject(uri, "GET", ConnectionGroup, Credential);
			request.Headers.Add("Translate", "f");

			response = request.GetResponse();

			responseStream = response.GetResponseStream();

			while((numBytes = responseStream.Read(bytes, 0, size)) > 0)
				stream.Write(bytes, 0, numBytes);
		}

		public void Post(string uri, string contentType, params Stream[] streams) 
		{
			const int size = 65536;
			byte[] bytes;
			int numBytes;
			HttpWebRequest request;
			WebResponse response;
			Stream requestStream;
			long contentLength;

			bytes = new byte[size];
			request = WebRequest.GetRequestObject(uri, "POST", ConnectionGroup, Credential);
			request.ContentType = contentType;

			contentLength = 0;
			foreach (Stream stream in streams) 
			{
				contentLength += stream.Length;
			}

			request.ContentLength = contentLength;

			requestStream = request.GetRequestStream();

			foreach (Stream stream in streams) 
			{
				while((numBytes = stream.Read(bytes, 0, size)) > 0)
					requestStream.Write(bytes, 0, numBytes);
			}
			response = request.GetResponse();
		}

		public NameValueCollection Options(string uri) 
		{
			WebRequest request;
			NameValueCollection headers;

			headers = new NameValueCollection();
			request = new WebRequest(ConnectionGroup);
			request.Request(uri, "OPTIONS", null, null, headers, Credential);

			return headers;
		}

		public string Subscribe(string uri, string notificationType, int depth, string callback, int lifeTime) 
		{
			NameValueCollection headers;
			
			headers = new NameValueCollection();
			headers.Add("Notification-type", notificationType);
			headers.Add("Depth", depth.ToString());
			headers.Add("Subscription-lifetime", lifeTime.ToString());
			headers.Add("Call-Back", callback);

			lock (webRequest) 
			{
				webRequest.Request(uri, "SUBSCRIBE", headers, null, headers, Credential);
			}

			return headers["Subscription-id"];
		}

		public void UnSubscribe(string uri, string subscriptionId) 
		{
			NameValueCollection headers;

			headers = new NameValueCollection();
			headers.Add("Subscription-ID", subscriptionId);

			lock (webRequest) {
				webRequest.Request(uri, "UNSUBSCRIBE", headers, Credential);
			}
		}

		public string[] Poll(string uri, params string[] subscriptionIds) 
		{
			NameValueCollection headers;
			XmlDocument doc;
			XmlNodeList list;
			string[] result;
			XmlNamespaceManager nsmgr;


			headers = new NameValueCollection();

			headers.Add("Subscription-ID", string.Join(",", subscriptionIds));

			lock (webRequest) 
			{
				doc = webRequest.Request(uri, "POLL", headers, _Credential);
			}

			nsmgr = new XmlNamespaceManager(doc.CreateNavigator().NameTable); 
			nsmgr.AddNamespace("b", "http://schemas.microsoft.com/Exchange/");
			nsmgr.AddNamespace("a", "DAV:");

			list = doc.SelectNodes("//b:subscriptionID/li[parent::*/parent::*/a:status[(.='HTTP/1.1 200 OK')]]", nsmgr);

			result = new string[list.Count];

			for (int i = 0; i<result.Length; i++) 
			{
				result[i] = list[i].InnerText;
			}

			return result;
		}
	}
}